---/eval live.load_module("player_dummy", "module")

local mod = ... or {}
local MODNAME = mod.MODNAME or "NONE"
local live = _G.live

if not live then
    live = {
        register_entity = core.register_entity,
        register_on_leaveplayer = function(_, func)
            core.register_on_leaveplayer(func)
        end,
        register_on_joinplayer = function(_, func)
            core.register_on_joinplayer(func)
        end,
    }
end

local function tell(...)
    core.chat_send_all(string.format(...))
end

local known_dummies = mod.known_dummies
if not mod.known_dummies then
    known_dummies = {}
    mod.known_dummies = known_dummies
end

--------------------------------------------------------------------------------

whosit = _G.whosit or {}
_G.whosit = whosit


MODNAME = 'whosit'
live.register_entity(
    ":" .. MODNAME .. ":dummy",
    {
        initial_properties = {
            visual = "mesh",
            mesh = "mobs_mime_plantlike.obj",
            textures = { "meh.png" },
            use_texture_alpha = true,
            hp_max = 1,
            physical = true,
            static_save = false,
        },
        on_activate = function(self, staticdata, dtime)
            local data = minetest.deserialize(staticdata)
            --tell("activate %s", dump(data))
            if data then
                if data.props then
                    self.object:set_properties(data.props)
                end
                -- self.nametag = data.nametag or ""
                -- self.object:set_nametag_attributes({
                --         text = self.nametag,
                --         color = "#FFFFFF",
                --         bgcolor = false,
                -- })
            end
        end,
        get_staticdata = function(self)
            local props = self.object:get_properties()
            local data = {
                props = props,
            }
            return minetest.serialize(data)
        end,
        on_step = function(self, dtime)
            --self.object:set_yaw(self.object:get_yaw() + 0.05)
            local bpm = 123
            local time = math.pi*self.time*(bpm/60)
            local rot = self.object:get_rotation()
            rot.y = rot.y + 0.03
            rot.x = math.sin(time)*0.1
            rot.z = math.cos(time*1.6)*0.05
            self.object:set_rotation(rot)
            self.time = (self.time or 0) + dtime

            local function dance1()
                local rot = function(bone, v)
                    local override = self.object:get_bone_override(bone)
                    if not override.rotation then
                        override.rotation = {}
                    end
                    override.rotation.interpolation = dtime*2
                    override.rotation.vec = v
                    self.object:set_bone_override(bone, override)
                end


                local angle = (math.pi/5)*math.sin(time)
                local angle2 = (math.pi/5)*math.sin(time/2)
                rot('Body', vector.new(angle,0,0))
                rot('Leg_Right', vector.new(-angle+(angle/10),0,0))
                rot('Leg_Left', vector.new(-angle-(angle/10),0,0))
                rot('Head', vector.new(-angle,angle2,0))
                rot('Arm_Right', vector.new(angle+math.pi/2,0,math.abs(angle2/2)))
                rot('Arm_Left', vector.new(angle+math.pi/2,0,-math.abs(angle2/2)))
            end


            local function dance2()
                local rot = function(bone, v)
                    local override = self.object:get_bone_override(bone)
                    if not override.rotation then
                        override.rotation = {}
                    end
                    override.rotation.interpolation = dtime*2
                    override.rotation.vec = v
                    self.object:set_bone_override(bone, override)
                end

                local bpm = 123
                local pi = math.pi
                local time = pi*self.time*(bpm/60)

                local angle = (pi/5)*(math.sin(time))
                local angle2 = (pi/5)*math.sin(time*2)
                local angle3 = (pi/5)*math.sin(time/2)
                local k = 1.5 -- [0.001,4]
                --local steep = math.atan(k*math.sin(time))/math.atan(k)
                local steep = math.sin(time)
                rot('Head', vector.new(angle3/4,0,angle))
                rot('Body', vector.new(0,0,angle/4))
                rot('Leg_Right', vector.new(0,angle/4,angle/4))
                rot('Leg_Left', vector.new(0,angle/4,angle/4))
                rot('Arm_Right', vector.new(pi+angle2,0,-pi/5+angle/1.5))
                rot('Arm_Left', vector.new(pi+angle2,0,pi/5+angle/1.5))
            end

            -- self.object:set_bone_override('Body', {})
            -- self.object:set_bone_override('Leg_Left', {})
            -- self.object:set_bone_override('Leg_Right', {})
            -- self.object:set_bone_override('Arm_Right', {})
            -- self.object:set_bone_override('Arm_Left', {})


            if true or string.find(self.object:get_properties().nametag, "Alias", 1, true) then
                dance1()
            else
                dance2()
            end

        end,
        on_death = function(self, killer)
            tell("killed")
        end,
    }
)

function whosit.make_clone(player)
    local pos = player:get_pos()
    if pos then
        local o = core.add_entity(pos, MODNAME .. ":dummy")
        local props = table.copy(player:get_properties())
        if (not props.nametag) or props.nametag == "" then
            props.nametag = player:get_player_name()
        end
        o:set_properties(props)

        -- nametag
        local attrs = table.copy(player:get_nametag_attributes())
        if (not attrs.text) or attrs.text == "" then
            attrs.text = player:get_player_name()
        end
        o:set_nametag_attributes(attrs)

        -- animation
        local range, speed, blend, loop = player:get_animation()
        o:set_animation(range, speed, blend, loop)

        -- yaw
        o:set_yaw(player:get_look_horizontal())
        return o
    end
end


-- live.register_on_leaveplayer(
--     MODNAME .. ":leave",
--     function(player, _timed_out)
--         local o = mod.make_clone(player)
--         local player_name = player:get_player_name()
--         known_dummies[player_name] = o
--     end
-- )


-- live.register_on_joinplayer(
--     MODNAME .. ":join",
--     function(player, last_login)
--         local player_name = player:get_player_name()
--         local dummy = known_dummies[player_name]
--         if dummy and dummy:is_valid() then
--             local pos = dummy:get_pos()
--             if pos then
--                 player:set_pos(pos)
--             end
--             dummy:remove()
--         end
--     end
-- )


-- live.register_tool(
--     MODNAME .. ":reloader",
--     {
--         on_use = function(itemstack, user, pointed_thing)
--             core.chat_send_all("player_dummy reloading...")
--             live.load_module("player_dummy", "module")
--         end,
--     }
-- )

local reload_interval = 2
local reload_time = 0
live.register_globalstep(
    "autoreload",
    function(dtime)
        reload_time = reload_time + dtime
        if reload_time > reload_interval then
            live.load_module("player_dummy", "module")
            reload_time = 0
        end
    end
)
