---/eval live.load_module("player_dummy", "module")

local mod = ... or {}
local MODNAME = mod.MODNAME or "NONE"
local live = _G.live

if not live then
    live = {
        register_entity = core.register_entity,
        register_on_leaveplayer = function(_, func)
            core.register_on_leaveplayer(func)
        end,
        register_on_joinplayer = function(_, func)
            core.register_on_joinplayer(func)
        end,
        safe_wrap = whosit.safe_wrap or function(f) return f end,
    }
end

local function tell(...)
    core.chat_send_all(string.format(...))
end

local known_dummies = mod.known_dummies
if not mod.known_dummies then
    known_dummies = {}
    mod.known_dummies = known_dummies
end

--------------------------------------------------------------------------------

whosit = _G.whosit or {}
_G.whosit = whosit
whosit[MODNAME] = mod

local pi = math.pi

local parameters = {
    -- `bone` is the name of the bone to rotate ('head')
    -- `sideways_angle` is an angle for animals that will look sideways, not directly

    ['petz:pony'] = {
        head_offset = vector.new(0,2,0),
        sideways_angle = pi/4,
    },
    default = {
        bone = 'head',
        head_offset = vector.new(0,0,0),
        sideways_angle = 0,
        yaw_limit = pi/2,
        pitch_limit = pi*0.1,
    },
}

-- get actual eye_pos of the player (including eye_offset)
local function player_get_eye_pos(player)
    local p_pos = player:get_pos()
    local p_eye_height = player:get_properties().eye_height
    p_pos.y = p_pos.y + p_eye_height
    local p_eye_pos = p_pos
    local p_eye_offset = vector.multiply(player:get_eye_offset(), 0.1)
    local yaw = player:get_look_horizontal()
    p_eye_pos = vector.add(p_eye_pos, vector.rotate_around_axis(p_eye_offset, {x=0,y=1,z=0}, yaw))
    return p_eye_pos
end

local function pony_look_at(self, at_pos)
    -- `self` is a pony reference
    -- `at_pos` is a point to look at
    local entity_name = self.name
    local params = parameters[entity_name] or parameters.default

    local function clamp(v, min_v, max_v)
        return math.min(math.max(v, min_v), max_v)
    end

    local pos = self.object:get_pos()

    local body_rotation = self.object:get_rotation()
    -- look direction
    local dir = vector.direction(pos, at_pos)
    -- `rotation` (vector) that will turn the head towards dir
    local dir_rot = vector.dir_to_rotation(dir) --, vector.new(1,0,0))
    -- make absolute look direction relative to body
    -- (so we just turn left/right instead of west/east)
    local rel_rot = dir_rot - body_rotation

    -- "head_turn" is jut rel_rot.y, but we want to make sure it's an
    -- angle in range [-pi,pi) or [-180,180) - where 0 is looking
    -- forward and 180 is looking back (we need this because luanti
    -- can return weird angles)
    local head_turn = ((rel_rot.y + 3.0*pi) % (2.0*pi)) - pi

    -- turn the head slightly sideways
    if head_turn < 0 then
        head_turn = head_turn + params.sideways_angle
    else
        head_turn = head_turn - params.sideways_angle
    end

    -- limit head rotations to some reasonable range, so it does not
    -- turn backwards
    local yaw_limit = params.yaw_limit
    local pitch_limit = params.pitch_limit
    head_turn = clamp(head_turn, -yaw_limit, yaw_limit)
    rel_rot.x = clamp(rel_rot.x, -pitch_limit, pitch_limit)

    local pitch = rel_rot.x
    local yaw = -head_turn
    local roll = rel_rot.z
    self.object:set_bone_override(params.bone, {rotation = { vec = vector.new(pitch, roll, yaw) }})
end


MODNAME = 'whosit'
live.register_entity(
    ":" .. MODNAME .. ":dummy2",
    {
        initial_properties = {
            visual = "mesh",
            mesh = "mobs_mime_plantlike.obj",
            textures = { "meh.png" },
            use_texture_alpha = true,
            hp_max = 1,
            physical = true,
            static_save = false,
        },
        on_activate = function(self, staticdata, dtime)
            local data = minetest.deserialize(staticdata)
            --tell("activate %s", dump(data))
            if data then
                if data.props then
                    self.object:set_properties(data.props)
                end
                -- self.nametag = data.nametag or ""
                -- self.object:set_nametag_attributes({
                --         text = self.nametag,
                --         color = "#FFFFFF",
                --         bgcolor = false,
                -- })
            end
        end,
        get_staticdata = function(self)
            local props = self.object:get_properties()
            local data = {
                props = props,
            }
            return minetest.serialize(data)
        end,
        on_step = function(self, dtime)
            -- self.object:set_yaw(self.object:get_yaw() + 0.1)
            -- local rot = self.object:get_rotation()
            -- rot.y = rot.y + 0.03
            -- self.object:set_rotation(rot)
            self.time = (self.time or 0) + dtime

            local pos = self.object:get_pos()
            local player = core.get_player_by_name('whosit')
            local p_pos = player:get_pos()

            self.object:set_yaw(self.object:get_yaw() + 0.6*dtime)

            local function dance()
                local function clamp(v, min_v, max_v)
                    return math.min(math.max(v, min_v), max_v)
                end

                local body_rotation = self.object:get_rotation()

                local rotate = function(bone, v)
                    local override = self.object:get_bone_override(bone)
                    if not override.rotation then
                        override.rotation = {}
                    end
                    override.rotation.interpolation = dtime*2
                    override.rotation.vec = v
                    self.object:set_bone_override(bone, override)
                end

                local bpm = 123
                local pi = math.pi
                local time = pi*self.time*(bpm/60)

                --local t_angle = (pi/5)*(math.sin(time))

                local dir = vector.direction(pos, p_pos)
                local dir_rot = vector.dir_to_rotation(dir) --, vector.new(1,0,0))
                local rel_rot = dir_rot - body_rotation
                local correction = math.pi/2

                local head_turn = ((rel_rot.y + 3.0*pi) % (2.0*pi)) - pi
                --print(head_turn)

                -- turn slightly (not that one)
                if head_turn < 0 then
                    head_turn = head_turn + pi/4
                else
                    head_turn = head_turn - pi/4
                end

                -- clamp
                head_turn = clamp(head_turn, -pi/2, pi/2)
                rel_rot.x = clamp(rel_rot.x, -pi*0.1, pi*0.1)

                local abs_rot = rel_rot - body_rotation
                -- local pitch = 0
                -- local yaw = angle
                -- local roll = 0

                local pitch = abs_rot.x
                local yaw = -head_turn
                local roll = abs_rot.z

                if true or math.ceil(self.time/4)%2 == 1 then
                    rotate('head', vector.new(pitch, roll, yaw))
                else
                    self.object:set_bone_override('head', {})
                end
                --rotate('head', rot)
            end

            local function reset()
                self.object:set_bone_override('head', {})
            end

            --dance()

            local player_eye_pos = player_get_eye_pos(player)
            pony_look_at(self, player_eye_pos, 'head', 0)

        end,
        on_death = function(self, killer)
            tell("killed")
        end,
    }
)

function mod.make_pony(player, other)
    local pos = player:get_pos()
    if not other then
        return
    end
    if pos then
        local o = core.add_entity(pos, MODNAME .. ":dummy2")
        local props = table.copy(other:get_properties())
        if (not props.nametag) or props.nametag == "" then
            props.nametag = player:get_player_name()
        end
        o:set_properties(props)

        -- nametag
        local attrs = table.copy(other:get_nametag_attributes())
        if (not attrs.text) or attrs.text == "" then
            attrs.text = player:get_player_name()
        end
        o:set_nametag_attributes(attrs)

        -- animation
        local range, speed, blend, loop = other:get_animation()
        o:set_animation(range, speed, blend, loop)

        -- yaw
        o:set_yaw(player:get_look_horizontal())
        return o
    end
end


-- live.register_on_leaveplayer(
--     MODNAME .. ":leave",
--     function(player, _timed_out)
--         local o = mod.make_clone(player)
--         local player_name = player:get_player_name()
--         known_dummies[player_name] = o
--     end
-- )


-- live.register_on_joinplayer(
--     MODNAME .. ":join",
--     function(player, last_login)
--         local player_name = player:get_player_name()
--         local dummy = known_dummies[player_name]
--         if dummy and dummy:is_valid() then
--             local pos = dummy:get_pos()
--             if pos then
--                 player:set_pos(pos)
--             end
--             dummy:remove()
--         end
--     end
-- )


-- live.register_tool(
--     MODNAME .. ":reloader",
--     {
--         on_use = function(itemstack, user, pointed_thing)
--             core.chat_send_all("player_dummy reloading...")
--             live.load_module("player_dummy", "module")
--         end,
--     }
-- )

local autoreload_job
local function autoreload(interval)
    interval = interval or 2


    live.safe_wrap(live.load_module)("player_dummy", "pony")
    autoreload_job = core.after(interval, autoreload)
end

function mod.start_autoreload()
    if not autoreload_job then
        live.log("autoreload started")
        autoreload()
    end
end

function mod.stop_autoreload()
    if autoreload_job then
        autoreload_job:cancel()
    end
end


-- local reload_interval = 2
-- local reload_time = 0
-- live.register_globalstep(
--     "autoreload",
--     function(dtime)
--         reload_time = reload_time + dtime
--         if reload_time > reload_interval then
--             live.load_module("player_dummy", "module")
--             reload_time = 0
--         end
--     end
-- )
